# -*- coding: utf-8 -*-

import sys
import xml.dom.minidom
import xmlrpclib
import re
from optparse import OptionValueError
from bkr.client.task_watcher import *
from bkr.client import BeakerCommand, BeakerWorkflow, BeakerJob, BeakerRecipeSet, BeakerRecipe

# CONSTANTS
DEFAULT_TIME_LIMIT = 86400
ERRATA_HTTP = "http://errata-xmlrpc.devel.redhat.com"
ERRATA_XMLRPC = "%s/errata/errata_service" % ERRATA_HTTP
ERRATA_NUMBER_EXTRACTOR = re.compile(r'\d{4}\:\d+')
TCP_PATH = "http://nest.test.redhat.com/mnt/qa/rhts/tc/%s.txt"
defaulttimelimit = 86400

#thx, jhutar
def uniquize(alist):
    """Make list member unique, keeping ordering"""
    alist_new = []
    for item in alist:
        if item not in alist_new:
            alist_new.append(item)
    return alist_new

# ET and RHTS use different formats for distro info
# this needs to be updated manually
distro_map = {
    'RHEL-3'      :   ("RHEL3-U9",),
    'RHEL-4'      :   ("RHEL4-U8",),
    'RHEL-4.7.Z'  :   ("RHEL4-U7",),
    'RHEL-5'      :   ("RHEL5-Server-U5", "RHEL5-Client-U5"),
    'RHEL-6'      :   ("RHEL6-6.0",),
    'RHEL-5.3.Z'  :   ("RHEL5-Server-U3", "RHEL5-Client-U3"),
    'RHEL-5.4.Z'  :   ("RHEL5-Server-U4", "RHEL5-Client-U4"),
    'RHEL-5.5.Z'  :   ("RHEL5-Server-U5", "RHEL5-Client-U5"),
    }

def already_warned():
    yield False
    while True: yield True

class Workflow_Errata(BeakerWorkflow):
    """Errata workflow to generate job to scheduler"""
    enabled = True
    debug = False
    doc = xml.dom.minidom.Document()

    def message(self, msg, type):
        if type == "DEBUG" and self.debug:
            print msg
        elif type == "NORMAL" and not self.errataTool:
            print msg
        elif type == "ETOOL" and self.errataTool:
            print msg
        elif type == "BOTH":
            print msg

    def et2rhts_map(self, et_name):
        """Maps Errata Tool distro representation to RHTS trees"""
        try:
            if et_name == "RHEL-2.1" and not self.errataTool:
                aw=already_warned()
                if not aw.next():
                    self.message("WARNING: Omitting 2.1 distro, RHTS doesn't support it", "NORMAL")
                return None
            else:
                return [rhts_name for rhts_name in distro_map[et_name] ]
        except KeyError:
            self.message("ERROR: There isn't a correct mapping for '%s' distro" % et_name, "BOTH")
            self.message("ERROR: You may use a -d option for explicit selection", "NORMAL")
            self.message("ERROR: Contact Petr Muller <pmuller@redhat.com> to add the mapping", "BOTH")
            sys.exit(2)

    def options(self):
        super(Workflow_Errata, self).options()
        self.parser.remove_option("--clients")
        self.parser.remove_option("--servers")
        self.parser.remove_option("--family")
        self.parser.remove_option("--distro")
        self.parser.add_option("-d", "--distro", dest="distro", metavar="DISTRO",
             action="append", default = [],help="Use distribution tree DISTRO")
        self.parser.add_option("--errata-tool", dest="errata_tool", action="store_true",
             default=False, help="""Do NOT use this option. Errata tool uses it.""")
        self.parser.add_option("--dnd", dest="dnd", action="store_true",
             default=False, help="""Do not attempt to downgrade in prepare stage.\
This option is contributed to Gary Gygax""")
        self.parser.add_option("--false-pass", dest="false_pass", action="store_true",
             default=False, help="""Report ALL results from single tests as PASS, so\
they don't spoil the result of recipe run. Please only use when you really\
know what it really does, so you won't miss a FAIL because of it.""")
        self.parser.add_option("--remove-package", dest="package_to_remove",
             action="append", default=[],
             help="""Before running all tests, RCW will perform rpm -e on these\
packages. Use with caution.""")
        self.parser.add_option("-Q", "--quick-run", dest="quickrun", action="store_true",
             help="Skip first batch of tests. This will cause only new packages to be\
tested. Compare stage will be skipped too, as there won't be anything\
else to compare to")
        self.parser.add_option("--norhtslib", dest="norhtslib", action="store_true",
             default=False, help="""Use old (nonrhtslibbed) variant of rcw pseudo-tests""")
        self.parser.add_option("-e", "--tests-for-errata", dest="test_errata",
             metavar="TEST_ERRATA", action="append", default=[],
             help="Run all tests relevant for all packages in TEST_ERRATA")
        self.parser.add_option("-E", "--update-errata", dest="update_errata",
             metavar="UPDATE_ERRATA", action="append", default=[],
             help="Update packages supplied as new in UPDATE_ERRATA")
        self.parser.add_option("-T", "--test-errata", dest="whole_errata",
             metavar="ERRATA", action="append", default=[],
             help="""Run all test for packages in ERRATA, and update to new packages\
 in ERRATA. This is equivalent to -e ERRATA -E ERRATA.""")
        self.parser.add_option("-l", "--list", dest="errata_list", action="store_true",
             help="List currently testable errata advisories")
        self.parser.usage = "%%prog %s [options]" % self.normalized_name

    def run(self, *args, **kwargs):
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)

        debug  = kwargs.get("debug", False)
        self.errataTool = kwargs.get("errata_tool", False)
        dryrun = kwargs.get("dryrun", False)
        wait = kwargs.get("wait", False)
	taskParams = kwargs.get("taskparam", [])

        # if we use -T errata, we can just append the selected errata
        # to 'update' and 'test' lists
        for errata in kwargs['whole_errata']:
            errata_number = ERRATA_NUMBER_EXTRACTOR.search(errata)
            if errata_number is not None:
                kwargs['update_errata'].append(errata_number.group(0))
                kwargs['test_errata'].append(errata_number.group(0))

        # if we work with errata, we need ET's xmlrpc server
        if kwargs['errata_list'] or kwargs['test_errata'] or kwargs['update_errata']:
            errata_server = xmlrpclib.ServerProxy(ERRATA_XMLRPC)

        # this just prints testable errata
        # those in NEED_QE, QE_READY, NEW, ON_QA or ON_RHNQA status
        STATUSES = ("NEED_QE", "QE_READY", "ON_QA", "ON_RHNQA", "NEW")
        if kwargs['errata_list']:
            errata_list = []
            for e in errata_server.errata.getRequestsByStatus(STATUSES):
                if e.has_key('fulladvisory') and e.has_key('synopsis'):
                    errata_list.append("%s %s" % (e['fulladvisory'], e['synopsis']))
            if len(errata_list) == 0:
                self.message("No errata found!", "NORMAL")
            else:
                self.message("%s" % "\n".join(errata_list), "NORMAL")
            sys.exit(0)

        # if we update errata, we'll store distros the errata is affecting
        # for later distro guess
        errata_releases = []

        errata_releases = uniquize(errata_releases)
        for release in errata_releases:
            self.message("%s distro was mapped as %s in RHTS" % (release[1], release[0]), "DEBUG")

        if not kwargs['distro']:
            for errata in kwargs['update_errata']:
                errata_packages = errata_server.errata.get_base_packages_rhts(errata)
                for errata_item in errata_packages:
                    releases = self.et2rhts_map(errata_item['rhel_version'])
                    if releases:
                        errata_releases.extend(releases)
            if errata_releases:
                while None in errata_releases:
                    errata_releases.remove(None)
                kwargs['distro'].extend(errata_releases)
                kwargs['distro'] = uniquize(kwargs['distro'])
                self.message("Errata driven distro choosing", "DEBUG")
        else:
            self.message("At least one distro should be specified", "BOTH")
            self.message("If you test errata with -T or -E then the workflow will \
try to choose automatically", "NORMAL")
            sys.exit(15)

        self.message("Distros:          %s" % kwargs['distro'], "DEBUG")

        # make list of packages for which all of the relevant tests will be run
        packages = kwargs['package']
        for errata in kwargs['test_errata']:
            erratapkgs = errata_server.errata.get_base_packages_rhts(errata)
            for errata_item in erratapkgs:
                packages.extend(errata_item['packages'])
        self.message("Packages:         %s" % packages, "DEBUG")

        kwargs['package'] = uniquize(packages)

        # get all tasks requested
        requestedTasks = self.getTasks(*args, **kwargs)
 
        params = ["TEST_PARAM_QA_TEST=YES"]

        if kwargs['dnd']:
            params.append("TEST_PARAM_DND=YES")

        if kwargs['update_errata']:
            errata_list = ",".join(uniquize([  ERRATA_NUMBER_EXTRACTOR.search(errata).group(0) for errata in  kwargs['update_errata'] ]))
            updated_errata_list = "TEST_PARAM_COMP_ERRATA=" + errata_list
            params.append(updated_errata_list)
            whiteboard = kwargs.pop("whiteboard")
            kwargs["whiteboard"] = "%s %s" % (whiteboard, kwargs['update_errata'])
            self.message ("Updated errata:   %s" % updated_errata_list, "DEBUG")

        pkglist = " ".join(packages)
        
        if kwargs['false_pass']:
            params.append("TEST_PARAM_FALSE_PASS=YES")
        else:
            params.append("TEST_PARAM_FALSE_PASS=NO")

        if kwargs['package_to_remove']:
            params.append("TEST_PARAM_REMOVE_ON_START=%s" % " ".join(kwargs['package_to_remove']))

        if kwargs['norhtslib']:
            appendix = ""
        else:
            appendix = "-rhtslib"

        if not kwargs['family'] and not kwargs['distro']:
            sys.stderr.write("No Family or Distro specified\n")
            sys.exit(1)

        distros = kwargs.pop('distro')

        # Create Job
        job = BeakerJob(*args, **kwargs)

        # Create Base Recipe
        recipeTemplate = BeakerRecipe()

        done = []

        for distro in distros:
            if distro in done:
                continue
            done.append(distro)
            if kwargs['arch']:
                # Take what the user requested but remove choices that don't apply.
                # RHEL5Client only supports i386 and x86_64, s390x invalid.
                arches = set(kwargs['arch']).intersection(set(self.getArches(distro=distro)))
            else:
                arches = self.getArches(distro=distro)
            for arch in arches:
                recipe = recipeTemplate.clone()
                # Add Distro Requirements
                temp = dict(kwargs)
                temp['distro'] = distro
                recipe.addBaseRequires(*args, **temp)
                arch_node = self.doc.createElement('distro_arch')
                arch_node.setAttribute('op', '=')
                arch_node.setAttribute('value', arch)
                recipe.addDistroRequires(arch_node)
        
                recipe.addTask('/distribution/install')
                recipe.addTask('/distribution/comparative/prepare%s' % appendix, taskParams=params)
                if not kwargs['quickrun']:
                    for task in requestedTasks:
                        if arch not in task['arches']:
                            recipe.addTask(task['name'], taskParams=params)
                else:
                    self.message('Quick run scheduled, first batch of tests will be skipped', 'BOTH')
                recipe.addTask('/distribution/comparative/upgrade%s' % appendix, taskParams=params)
                for task in requestedTasks:
                    if arch not in task['arches']:
                        recipe.addTask(task['name'], taskParams=params)
                if not kwargs['quickrun']:
                    recipe.addTask('/distribution/comparative/gather%s' % appendix, 
                                   taskParams=params)

                job.addRecipe(recipe)

        # jobxml
        jobxml = job.toxml(**kwargs)

        if debug:
            print jobxml

        self.set_hub(username, password)
        submitted_jobs = []
        failed = False

        if not dryrun:
            try:
                submitted_jobs.append(self.hub.jobs.upload(jobxml))
            except Exception, ex:
                failed = True
                print ex
        if not dryrun:
            print "Submitted: %s" % submitted_jobs
            if wait:
                watch_tasks(self.hub, submitted_jobs)
            if failed:
                sys.exit(1)
