# Copyright (C) 2011 Jan Stancek <jstancek@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
__author__ = """Copyright Jan Stancek 2011"""

"""
cmd_workflow_kt1auto - runs workflow-kernel with parameters deducted from
nvr_tables obtained from first available source:
- command line parameter: --kt1auto-tables=<path/url>
- env. variable: KT1AUTO_TABLES=<path/url>
- url defined in beaker config: KT1AUTO_TABLES = "<path/url>"
- hardcoded git archive vspec
- hardcoded http URL
"""

TABLES_FILENAME = 'kt1_nvr_tables.json'
TABLES_GIT_ARCHIVE = 'git://pkgs.devel.redhat.com/tests/kernel master:workflows/kt1auto'
TABLES_GIT_URL = 'http://pkgs.devel.redhat.com/cgit/tests/kernel/plain/workflows/kt1auto/kt1_nvr_tables.json'


import re
import os
import sys
import time
import urllib2
import subprocess
import tempfile
import shutil
import shlex
import copy
import pipes

from bkr.client import conf
from bkr.client.commands.cmd_workflow_kernel import Workflow_Kernel

try:
    import simplejson as json
except ImportError:
    import json


def log(msg, *args, **kwargs):
        msg = (msg + '\n') % (args)
        sys.stderr.write(msg)


def parse_nvr(somestring):
    somestring = somestring.strip()
    pattern = re.compile('^(.+)-(.+)-(.+)$')
    match = pattern.search(somestring)
    if match:
        return match.group(1), match.group(2), match.group(3)
    else:
        return None


def compare_lists(list1, list2):
    i = 0
    while i < len(list1) and i < len(list2):
        part1 = list1[i]
        part2 = list2[i]

        if part1.isdigit():
            if part2.isdigit():
                part1 = int(part1)
                part2 = int(part2)
            else:
                return 1
        else:
            if part2.isdigit():
                return -1

        if part1 < part2:
            return -1
        elif part1 > part2:
            return 1
        i = i + 1

    if len(list1) > len(list2):
        return 1
    elif len(list1) < len(list2):
        return -1
    else:
        return 0


def compare_nvr(nvr1str, nvr2str):
    nvr1 = parse_nvr(nvr1str)
    nvr2 = parse_nvr(nvr2str)

    if nvr1 and nvr2:
        pass
    else:
        return None

    name1, ver1, rel1 = nvr1
    name2, ver2, rel2 = nvr2

    ver1 = ver1.split('.')
    ver2 = ver2.split('.')

    rel1 = rel1.split('.')
    rel2 = rel2.split('.')

    list1 = ver1 + rel1
    list2 = ver2 + rel2

    return compare_lists(list1, list2)


def nvr_in_range(nvr, nvr_low, nvr_high):
    if (compare_nvr(nvr_low, nvr) <= 0 and
        compare_nvr(nvr_high, nvr) >= 0):
        return True
    else:
        return False


def get_remote_file(someurl):
    RETRIES = 3
    RETRY_DELAY = 10
    i = 0
    data = None

    while data == None and i < RETRIES:
        try:
            usock = urllib2.urlopen(someurl)
            data = usock.read()
            if data.find('<meta http-equiv="refresh"') > 0:
                usock = urllib2.urlopen(someurl)
                data = usock.read()
        except (urllib2.URLError, urllib2.HTTPError), e:
            data = None
            log('error: get_remote_file: %s', str(e))

        if data == None:
            time.sleep(RETRY_DELAY)
            log('Trying again to get: %s', someurl)
            i += 1
    return data


def load_kt1_nvr_tables_git(gitvspec):
    ret = None
    temp_dir = tempfile.mkdtemp()
    git_cmd = 'git archive --remote=' + gitvspec + ' | tar x'
    subprocess.call(git_cmd, cwd=temp_dir, shell=True)
    tables_file = os.path.join(temp_dir, TABLES_FILENAME)
    if os.path.exists(tables_file):
        ret = load_kt1_nvr_tables_file(tables_file)
    shutil.rmtree(temp_dir)
    return ret


def load_kt1_nvr_tables_url(someurl):
    data = get_remote_file(someurl)
    ret = None
    if data:
        ret = json.loads(data)
    return ret


def load_kt1_nvr_tables_file(filename):
    ret = None
    f = open(filename, 'r')
    ret = json.load(f)
    f.close()
    return ret


def get_from_tables(kt1_nvr_tables, table_name, nvr, only_first=True):
    ret = []
    default_value = None
    if table_name in kt1_nvr_tables:
        table = kt1_nvr_tables[table_name]
        if len(table) > 0 and table[0][0] == 'default':
            default_value = table[0][2]
            table = table[1:]
        for some_tuple in table:
            nvr_low = some_tuple[0]
            nvr_high = some_tuple[1]

            if len(some_tuple) == 3:
                value = some_tuple[2]
            else:
                # if there are multiple values join them to single string
                value = ''
                for v in some_tuple[2:]:
                    value = value + str(v)
            if nvr_high == "":
                nvr_high = nvr_low
            if nvr_in_range(nvr, nvr_low, nvr_high):
                ret.append(value)
                if only_first:
                    break
    if only_first and len(ret) == 0:
        ret.append(default_value)
    return ret


class Workflow_KT1Auto(Workflow_Kernel):
    """Kernel workflow to generate job to scheduler"""
    enabled = True

    def options(self):
        super(Workflow_KT1Auto, self).options()
        self.wkp_ignore = []

        self.parser.add_option(
            "--kt1auto-tables",
            dest="kt1auto_tables",
            default=None,
            help="NVR tables for distro/family/tag/kvm/etc."
            )
        self.wkp_ignore.append('kt1auto-tables')

        self.parser.add_option(
            "--tag-bkr-job",
            dest="tag_bkr_job",
            default=False,
            action="store_true",
            help="Tag beaker job - applies product and retention tag"
            )
        self.wkp_ignore.append('tag-bkr-job')

        self.parser.add_option(
            "--show-wkparams",
            dest="wk_params",
            default=False,
            action="store_true",
            help="Show parameters for (parent) kernel-workflow"
            )
        self.wkp_ignore.append('show-wkparams')

        self.parser.add_option(
            "--dryrun-kt1auto",
            dest="dryrun_kt1auto",
            default=False,
            action="store_true",
            help="If True, parent workflow wont be called"
            )
        self.wkp_ignore.append('dryrun-kt1auto')

        self.parser.add_option(
            "--no-tables",
            dest="no_tables",
            default=False,
            action="store_true",
            help="If True ignore nvr tables completely"
            )
        self.wkp_ignore.append('no-tables')

        self.parser.usage = '%%prog %s [options]' % self.normalized_name

        # we have to backup defaults here, because they will get overwritten
        # on subsequent calls
        self.parser_defaults = copy.deepcopy(self.parser.defaults)

    def _get_tables_location(self, *args, **kwargs):
        tables_param = kwargs.get('kt1auto_tables', None)
        if tables_param:
            return tables_param

        tables_env = os.environ.get('KT1AUTO_TABLES', None)
        if tables_env:
            return tables_env

        tables_config = conf.get('KT1AUTO_TABLES', None)
        if tables_config:
            return tables_config

        return None


    def _load_kt1_nvr_tables(self, *args, **kwargs):
        kt1_nvr_tables = None
        tables_location = self._get_tables_location(*args, **kwargs)

        if tables_location:
            if tables_location.startswith('http'):
                kt1_nvr_tables = load_kt1_nvr_tables_url(tables_location)
            elif tables_location.startswith('git'):
                kt1_nvr_tables = load_kt1_nvr_tables_git(tables_location)
            else:
                kt1_nvr_tables = load_kt1_nvr_tables_file(tables_location)
        else:
            kt1_nvr_tables = load_kt1_nvr_tables_git(TABLES_GIT_ARCHIVE)
            if not kt1_nvr_tables:
                kt1_nvr_tables = load_kt1_nvr_tables_url(TABLES_GIT_URL)

        return kt1_nvr_tables


    def process_separate_tables(self, *args, **kwargs):
        """ process_separate_tables - old style of storing workflow parameters,
            where each parameter was in separate table
        """

        nvr = self.nvr
        kt1_nvr_tables = self.kt1_nvr_tables

        wtype = 'KernelTier1'
        kwargs['type'] = wtype

        distro = kwargs.get('distro', None)
        if not distro:
            distro = get_from_tables(kt1_nvr_tables, 'distro',
                                     nvr, only_first=True)[0]
            kwargs['distro'] = distro

        tag = None
        if not distro:
            tag = kwargs.get('tag', None)
            if not tag:
                tag = get_from_tables(kt1_nvr_tables, 'distro_tag',
                                      nvr, only_first=True)[0]
                kwargs['tag'] = [tag]

        family = kwargs.get('family', None)
        if not family:
            family = get_from_tables(kt1_nvr_tables, 'distro_family',
                                     nvr, only_first=True)[0]
            kwargs['family'] = family

        kvm = kwargs.get('kvm', None)
        if not kvm:
            kvm = get_from_tables(kt1_nvr_tables, 'kvm',
                                  nvr, only_first=True)[0]
            kwargs['kvm'] = kvm

        ndump = kwargs.get('ndump', None)
        kdump = kwargs.get('kdump', None)
        if not ndump and not kdump:
            ndump = get_from_tables(kt1_nvr_tables, 'ndump',
                                    nvr, only_first=True)[0]
            if ndump:
                kdump = False
                kwargs['ndump'] = True
            else:
                kdump = True
                kwargs['kdump'] = True

        repos = kwargs.get('repo', None)
        if not repos:
            kwargs['repo'] = []

        supp_repo = get_from_tables(kt1_nvr_tables, 'supp_repo',
                                    nvr, only_first=True)[0]
        if supp_repo:
            if not supp_repo in kwargs['repo']:
                kwargs['repo'].append(supp_repo)
        repos = kwargs['repo']

        product = None
        retention_tag = None

        tag_bkr_job = kwargs.get('tag_bkr_job', False)
        if tag_bkr_job:
            product = kwargs.get('product', None)
            if not product:
                product = get_from_tables(kt1_nvr_tables, 'product',
                                          nvr, only_first=True)[0]
                kwargs['product'] = product

            retention_tag = kwargs.get('retention_tag', None)
            if product:
                rtag_from_tables = True
                if retention_tag != None and retention_tag.lower() != 'scratch':
                    rtag_from_tables = False
                if rtag_from_tables:
                    retention_tag = get_from_tables(kt1_nvr_tables, 'retention_tag',
                                                    nvr, only_first=True)[0]
                kwargs['retention_tag'] = retention_tag

        all_args = []
        all_args.append('--type=KernelTier1')
        if nvr:
            all_args.append('--nvr=%s' % (nvr))
        if distro:
            all_args.append('--distro=%s' % (distro))
        if family:
            all_args.append('--family=%s' % (family))
        if tag:
            all_args.append('--tag=%s' % (tag))
        if kvm:
            all_args.append('--kvm')
        if ndump:
            all_args.append('--ndump')
        if kdump:
            all_args.append('--kdump')
        for repo in repos:
            all_args.append('--repo=%s' % (repo))
        if product:
            all_args.append('--product=%s' % (product))
        if retention_tag:
            all_args.append('--retention_tag=%s' % (retention_tag))

        args_to_ignore = ['type', 'nvr', 'distro', 'family', 'tag',
                          'kvm', 'ndump', 'kdump', 'repo', 'product','retention_tag']
        for arg in sys.argv[2:]:
            add_to_cmd = True
            for pi in args_to_ignore:
                if arg.find('--%s' % (pi)) != -1:
                    add_to_cmd = False
                    break
            if add_to_cmd:
                all_args.append(arg)

        return all_args, args, kwargs

    def process_combined_params_tables(self, *args, **kwargs):
        """ process_combined_params_tables - this is the new style of storing workflow parameters
            all params are in as single entry, consisting of tables:
            workflow_params_first_matching
            additional_workflow_params_all_matching
        """

        params = get_from_tables(self.kt1_nvr_tables,
                                       'workflow_params_first_matching',
                                       self.nvr, only_first=True)[0]

        additional_params = get_from_tables(self.kt1_nvr_tables,
                                       'additional_workflow_params_all_matching',
                                       self.nvr, only_first=False)

        for p in additional_params:
            params += ' %s' % (p)

        args_array = shlex.split(params)
        args_array.extend(sys.argv[2:])

        # restore parser defaults
        self.parser.defaults = self.parser_defaults
        cmd, cmd_opts, cmd_args = self.parser.parse_args(args=args_array, values=None)
        pargs = cmd_args
        pkwargs = cmd_opts.__dict__

        tag_bkr_job = kwargs.get('tag_bkr_job', False)
        if not tag_bkr_job:
            pkwargs['product'] = None
            pkwargs['retention_tag'] = None
            self.wkp_ignore.append('product')
            self.wkp_ignore.append('retention_tag')

        return args_array, pargs, pkwargs


    def process_tables(self, *args, **kwargs):
        exact_params = get_from_tables(self.kt1_nvr_tables,
                                       'workflow_params_first_matching',
                                       self.nvr, only_first=True)[0]
        if exact_params != None:
            return self.process_combined_params_tables(*args, **kwargs)
        else:
            return self.process_separate_tables(*args, **kwargs)


    def print_wk_params(self, all_args):
        cmd = ''
        for arg in all_args:
            add_to_cmd = True
            for pi in self.wkp_ignore:
                if arg.find('--%s' % (pi)) != -1:
                    add_to_cmd = False
                    break
            if add_to_cmd:
                cmd += '%s%s' % (' \\\n', pipes.quote(arg))
        print 'bkr workflow-kernel', cmd


    def run(self, *args, **kwargs):
        self.nvr = kwargs.get('nvr', None)
        if not self.nvr:
            sys.stderr.write('No nvr parameter found\n')
            sys.exit(3)

        self.kt1_nvr_tables = self._load_kt1_nvr_tables(*args, **kwargs)
        if not self.kt1_nvr_tables:
            sys.stderr.write('Error: No usable kt1 nvr tables found\n')
            sys.exit(2)

        no_tables = kwargs.get('no_tables')
        if no_tables:
            all_args = sys.argv[2:]
        else:
            all_args, args, kwargs = self.process_tables(*args, **kwargs)

        wk_params = kwargs.get('wk_params')
        if wk_params:
            self.print_wk_params(all_args)

        dryrun_kt1auto = kwargs.get('dryrun_kt1auto')
        if not dryrun_kt1auto:
            super(Workflow_KT1Auto, self).run(*args, **kwargs)



