# -*- coding: utf-8 -*-

from bkr.client.task_watcher import *
from bkr.client import BeakerCommand, BeakerWorkflow, BeakerJob, BeakerRecipeSet, BeakerRecipe
from optparse import OptionValueError
import os
import sys
import xml.dom.minidom
import snake.client, pykickstart, re
import pykickstart.version
import pykickstart.errors

def format_test_params(params):
    formatter = lambda k, v: "TEST_PARAM_%s=%s" % (k.upper(), v)
    return [formatter(*p.split("=", 1)) for p in params]

class Workflow_Snake(BeakerWorkflow):
    """Snake workflow"""
    enabled = True
    doc = xml.dom.minidom.Document()

    def options(self):
        super(Workflow_Snake, self).options()

        snake_server = os.getenv("SNAKESERVER", "").split(":")
        snake_port = len(snake_server) > 1 and int(snake_server[1]) or 2903
        snake_server = snake_server[0]

        self.parser.add_option(
            "--template",
            default="",
            help="SNAKE template to build kickstart from"
            )
        self.parser.add_option(
            "--snakeversion",
            default="",
            help="Specify a kickstart version (optional)"
            )
        self.parser.add_option(
            "--snakeserver",
            default=snake_server,
            help="SNAKE server to connect to (SNAKESERVER env by default)"
            )
        self.parser.add_option(
            "--snakeport",
            default=snake_port,
            type=int,
            help="SNAKE port to connect to (optional)"
            )
        self.parser.add_option(
            "--suppress-args",
            action="store_true",
            default=False,
            dest="suppress_args",
            help="Suppress kernel options provided by snake"
            )
        self.parser.add_option(
            "--testparam",
            action="append",
            default=[],
            help="Set task params 'name=value' as 'TEST_PARAM_NAME=value'"
            )
        self.parser.usage = "%%prog %s [options]" % self.normalized_name

    def getTasks(self, *args, **kwargs):
        snake_task = "/distribution/install/" + kwargs["template"]
        kwargs.setdefault('task', [])

        if snake_task not in kwargs['task']:
            kwargs['task'].insert(0, snake_task)

        return super(Workflow_Snake, self).getTasks(*args, **kwargs)

    def run(self, *args, **kwargs):
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)

        debug  = kwargs.get("debug", False)
        dryrun = kwargs.get("dryrun", False)
        wait = kwargs.get("wait", False)
	family = kwargs.get("family", None)
	distro = kwargs.get("distro", None)
	arches = kwargs.get("arches", [])
	taskParams = kwargs.get("taskparam", [])
	testParams = kwargs.get("testparam", [])

        # get snake options
        use_args = not kwargs.get("suppress_args", False)
        snakeTemplate = kwargs.get("template", None)
        bootArgs = kwargs.pop("kernel_options", None)
        snakeVersion = kwargs.get("snakeversion", None)
        snakeServer = kwargs.get("snakeserver", None)
        snakePort = kwargs.get("snakeport", 2903)

        if not family and not distro:
            sys.stderr.write("No Family or Distro specified\n")
            sys.exit(1)

        if not arches:
            sys.stderr.write("No arches specified, you must specify at least one\n")
            sys.exit(1)

        if not snakeServer or not snakePort:
            sys.stderr.write("You must specify snake server and port to run snake workflow\n")
            sys.exit(1)

        if not snakeTemplate:
            sys.stderr.write("You must specify snake template to run snake workflow\n")
            sys.exit(1)

        # get all tasks requested
        requestedTasks = self.getTasks(*args, **kwargs)

        # Add kickstart
        snake_server = snake.client.connect(snakeServer, snakePort)
        snake_args = {}
        # add boot args
        if bootArgs is None:
            bootArgs = ''
        if use_args and "kernel_args.generate" in snake_server.system.listMethods():
            argdata = dict()
            argdata['cmdline'] = bootArgs
            argdata['ks'] = snakeTemplate
            bootArgs = snake_server.kernel_args.generate(argdata)
        kwargs['kernel_options'] = bootArgs
        
        # make ksmeta from taskparam
        try:
            snake_args["ksmeta"] = dict(map(lambda x: x.split('=', 1), testParams + taskParams))
        except:
            sys.stderr.write("Every task param has to have value\n")
            sys.exit(1)

        if kwargs.get("variant",None):
            snake_args["ksmeta"]["variant"] = kwargs.get("variant")

        # If no kickstart snakeversion was specified ... let's guess
        if not snakeVersion:
            # Try to guess using family
            if family:
                # De-studly cap the family string (e.g.
                # "RedHatEnterpriseLinuxServer5" -> "Red Hat Enterprise Linux
                # Server 5") so that pykickstart can parse
                    try:
                        snakeVersion = pykickstart.version.stringToVersion(re.sub("([A-Z0-9])", " \\1", family).strip())
                    except pykickstart.errors.KickstartVersionError:
                        snakeVersion = "devel"
            # Try to guess using distro
            elif distro:
                rhel_match = re.match("RHEL-?([3-9]).*", distro, re.I)
                fedora_match = re.match("(FC?)-?([0-9]+).*", distro, re.I)
                if rhel_match:
                    snakeVersion = "rhel%s" % rhel_match.group(1)
                elif fedora_match:
                    snakeVersion = "%s%s" % (fedora_match.group(1).lower(), fedora_match.group(2))
                else:
                    snakeVersion = "devel"
            # Default to 'devel'
            else:
                snakeVersion = "devel"
        snake_args["version"] = snakeVersion

        # Create Job
        job = BeakerJob(*args, **kwargs)

        # convert params to TEST_PARAM_var=val
        taskParams = map(lambda x: "TEST_PARAM_%s=%s"%(x.split('=')[0].upper(),  x.split('=', 1)[1]), taskParams)

        if(len(taskParams) > 0):
            sys.stderr.write("DEPRECATION WARNING: Use testparam for conversion to 'TEST_PARAM_NAME=value', taskparam won't change parameters in the future!\n")

        taskParams = format_test_params(testParams) + taskParams

        # Add Host Requirements
        for arch in arches:
            snake_args["ksmeta"]["arch"] = arch
            # Create Base Recipe
            recipeTemplate = BeakerRecipe()
            # get kickstart and add it to recipes
            kickstart = snake_server.kickstart.generate(snakeTemplate, snake_args)
            recipeTemplate.addKickstart(kickstart)
            # Add Distro Requirements
            recipeTemplate.addBaseRequires(*args, **kwargs)

            arch_node = self.doc.createElement('distro_arch')
            arch_node.setAttribute('op', '=')
            arch_node.setAttribute('value', arch)
            if self.multi_host:
                recipeSet = BeakerRecipeSet()
                for i in range(self.n_servers):
                    recipeSet.addRecipe(self.processTemplate(recipeTemplate, 
                                                             requestedTasks,
                                                             taskParams=taskParams,
                                                             distroRequires=arch_node, 
                                                             role='SERVERS',
                                                             arch=arch,
                                                             **kwargs))
                for i in range(self.n_clients):
                    recipeSet.addRecipe(self.processTemplate(recipeTemplate, 
                                                             requestedTasks,
                                                             taskParams=taskParams,
                                                             distroRequires=arch_node, 
                                                             role='CLIENTS',
                                                             arch=arch,
                                                             **kwargs))
                job.addRecipeSet(recipeSet)
            else:
                job.addRecipe(self.processTemplate(recipeTemplate,
                                                   requestedTasks,
                                                   taskParams=taskParams,
                                                   distroRequires=arch_node,
                                                   arch=arch,
                                                   **kwargs))

        # jobxml
        jobxml = job.toxml(**kwargs)

        if debug:
            print jobxml

        self.set_hub(username, password)
        submitted_jobs = []
        failed = False

        if not dryrun:
            try:
                submitted_jobs.append(self.hub.jobs.upload(jobxml))
            except Exception, ex:
                failed = True
                print ex
        if not dryrun:
            print "Submitted: %s" % submitted_jobs
            if wait:
                watch_tasks(self.hub, submitted_jobs)
            if failed:
                sys.exit(1)
